/*                     __                                               *\
**     ________ ___   / /  ___     Scala API                            **
**    / __/ __// _ | / /  / _ |    (c) 2003-2009, LAMP/EPFL             **
**  __\ \/ /__/ __ |/ /__/ __ |    http://scala-lang.org/               **
** /____/\___/_/ |_/____/_/ | |                                         **
**                          |/                                          **
\*                                                                      */

// $Id: List.scala 17071 2009-02-10 09:32:05Z rytz $


package /*scala*/local.collection.mutable

/*-enable-unique*/
import scala.annotation.{transient, unique}

//import scala.collection.mutable.ListBuffer
import Predef._

/** A class representing an ordered collection of elements of type
 *  <code>a</code>. This class comes with two implementing case
 *  classes <code>scala.Nil</code> and <code>scala.::</code> that
 *  implement the abstract members <code>isEmpty</code>,
 *  <code>head</code> and <code>tail</code>.
 *
 *  @author  Martin Odersky and others
 *  @version 1.0, 16/07/2003
 */
@transient
sealed abstract class MyList[+A] extends /*Seq[A] with*/ Product {

  /** Returns true if the list does not contain any elements.
   *  @return <code>true</code>, iff the list is empty.
   */
  /*override*/ def isEmpty: Boolean

  /** Returns this first element of the list.
   *
   *  @return the first element of this list.
   *  @throws Predef.NoSuchElementException if the list is empty.
   */
  def head: A

  /** Result of comparing <code>length</code> with operand <code>l</code>.
   *  returns <code>x</code> where
   *  <code>x &lt; 0</code>    iff    <code>this.length &lt; l</code>
   *  <code>x == 0</code>   iff    <code>this.length == l</code>
   *  <code>x &gt; 0</code>    iff    <code>this.length &gt; that</code>.
   *
   *  This method is used by matching streams against right-ignoring (...,_*) patterns.
   *  
   *  This method does not call <code>List.length</code>, it works for <code>O(l)</code>,
   *  not for <code>O(length)</code>.
   */
  /*override*/ def lengthCompare(l: Int): Int = {
    if (isEmpty) 0 - l
    else if (l <= 0) 1
    else tail.lengthCompare(l - 1)
  }

  /** Returns this list without its first element.
   *
   *  @return this list without its first element.
   *  @throws Predef.NoSuchElementException if the list is empty.
   */
  def tail: MyList[A]

  /** Returns the number of elements in the list.
   *
   *  @return the number of elements in the list.
   */
  def length: Int = {
/*
    var these = this
    var len = 0
    while (!these.isEmpty) {
      len += 1
      these = these.tail
    }
    len
*/
    if (this.isEmpty) 0
    else this.tail.length
  }

  /** Returns the list without its <code>n</code> first elements.
   *  If this list has less than <code>n</code> elements, the empty list is returned.
   *
   *  @param n the number of elements to drop.
   *  @return the list without its <code>n</code> first elements.
   */
  /*override*/ def drop(n: Int): MyList[A] = {
/*
    var these = this
    var count = n
    while (!these.isEmpty && count > 0) {
      these = these.tail
      count -= 1
    }
    these
*/
    if (this.isEmpty || n == 0) this
    else this.tail.drop(n-1)
  }

  /** Returns the <code>n</code>-th element of this list. The first element
   *  (head of the list) is at position 0.
   *
   *  @param n index of the element to return
   *  @return  the element at position <code>n</code> in this list.
   *  @throws Predef.NoSuchElementException if the list is too short.
   */
  def apply(n: Int): A = drop(n).head

  /** Returns the elements in the list as an iterator
   *
   *  @return an iterator on the list elements.
   */
  /*override*/ def elements: Iterator[A] = new Iterator[A] {
    var these = MyList.this
    def hasNext: Boolean = !these.isEmpty
    def next: A =
      if (!hasNext)
        throw new NoSuchElementException("next on empty Iterator")
      else {
        val result: A@unique = these.head; these = these.tail; result
      }
//    override def toList: MyList[A] = these
  }

  /** Overrides the method in Iterable for efficiency.
   *
   *  @return  the list itself
   */
//  override def toList: MyList[A] = this

  /** A list consisting of all elements of this list in reverse order.
   */
  /*override*/ def reverse: MyList[A] = {
/*
    var result: MyList[A] = new Nil
    var these = this
    while (!these.isEmpty) {
      result = Cons(these.head, result)
      these = these.tail
    }
    result

*/
    if (this.isEmpty) new Nil
    else Cons(this.head, this.tail.reverse)
  }

}

/** The empty list.
 *
 *  @author  Martin Odersky
 *  @version 1.0, 15/07/2003
 */
@SerialVersionUID(0 - 8256821097970055419L)
@transient case class Nil() extends MyList[Nothing] {
  override def isEmpty = true
  def head: Nothing =
    throw new NoSuchElementException("head of empty list")
  def tail: MyList[Nothing] =
    throw new NoSuchElementException("tail of empty list")
}

/** A non empty list characterized by a head and a tail.
 *
 *  @author  Martin Odersky
 *  @version 1.0, 15/07/2003
 */
@SerialVersionUID(0L - 8476791151983527571L)
@transient final case class Cons[B](private var hd: B, var tl: MyList[B]) extends MyList[B] {
  def head : B = hd
  def tail : MyList[B] = tl
  override def isEmpty: Boolean = false
}
